import re

from pongpy.definitions import ATK_SIZE, DEF_SIZE
from pongpy.interfaces.team import Team
from pongpy.models.game_info import GameInfo
from pongpy.models.pos import Pos
from pongpy.models.state import State, TeamState

DMZ_SIZE = 8
SADDEN_DEATH_DURATION = 400
MAX_NAME_LENGTH = 16
NAME_PATTERN = re.compile(r'^[\w ]+$', re.A)


class TeamManager:
    game_info: GameInfo
    atk_pos: Pos
    def_pos: Pos
    score: int

    def __init__(self, gi: GameInfo, team: Team, reversed: bool):
        validate_name(team.name)
        self.team = team
        self.game_info = gi
        self.score = 0
        if not reversed:
            self.atk_pos = Pos(gi.width // 12 * 5, gi.height // 2)
            self.def_pos = Pos(gi.width // 12 * 1, gi.height // 2)
        else:
            self.atk_pos = Pos(gi.width // 12 * 7, gi.height // 2)
            self.def_pos = Pos(gi.width // 12 * 11, gi.height // 2)

    def update(self, state: State):
        # ゲームが長引いたらだんだん近づける
        direction = -1 if state.is_right_side else 1
        time_pos = direction * 0.1 if state.time % SADDEN_DEATH_DURATION == 1 else 0
        # これ以上は近づけない
        if self.game_info.width // 2 - DMZ_SIZE < self.atk_pos.x < self.game_info.width // 2 + DMZ_SIZE:
            time_pos = 0

        atk_action = self.team.atk_action(info=self.game_info, state=state)
        assert isinstance(atk_action, (int, float)), 'atk_action の返り値が数値ではありません'
        assert -self.game_info.atk_return_limit <= atk_action <= self.game_info.atk_return_limit, 'atk_action の返り値が大きすぎます'
        self.atk_pos = Pos(self.atk_pos.x + time_pos, self._move_y(atk_action, self.atk_pos, ATK_SIZE))
        def_action = self.team.def_action(info=self.game_info, state=state)

        assert isinstance(atk_action, (int, float)), 'def_action の返り値が数値ではありません'
        assert -self.game_info.def_return_limit <= def_action <= self.game_info.def_return_limit, 'def_action の返り値が大きすぎます'
        self.def_pos = Pos(self.def_pos.x + time_pos * 5, self._move_y(def_action, self.def_pos, DEF_SIZE))

    def add_score(self):
        self.score += 1

    def _move_y(self, y_delta: float, pos: Pos, size: int) -> int:
        if y_delta + pos.y - size // 2 < 0:
            return 0 + size // 2
        if y_delta + pos.y + size // 2 > self.game_info.height:
            return self.game_info.height - size // 2
        return int(y_delta + pos.y)

    @property
    def state(self) -> TeamState:
        return TeamState(atk_pos=self.atk_pos, def_pos=self.def_pos, score=self.score)

    @property
    def score_label(self) -> str:
        return f'{self.team.name}: {self.score}'


def validate_name(name: str):
    assert NAME_PATTERN.match(name), 'チーム名に ascii 以外の文字列が含まれてます'
    assert 0 < len(name) <= MAX_NAME_LENGTH, 'チーム名の長さが不正です'


def _isascii(s: str):
    return len(s) == len(s.encode())



from pongpy.interfaces.team import Team
from pongpy.models.game_info import GameInfo
from pongpy.models.state import State


class FollowTeam(Team):
    """
    現在のボールの位置似合わせて動くサンプル。
    """
    @property
    def name(self) -> str:
        return 'follow team'

    def atk_action(self, info: GameInfo, state: State) -> int:
        direction = (state.ball_pos.y - state.mine_team.atk_pos.y) > 0
        return info.atk_return_limit if direction else -info.atk_return_limit

    def def_action(self, info: GameInfo, state: State) -> int:
        direction = (state.ball_pos.y - state.mine_team.def_pos.y) > 0
        return info.def_return_limit if direction else -info.def_return_limit



from logging import getLogger

import pyxel

from pongpy.controllers.board import Board
from pongpy.definitions import PADDING, HEIGHT, WIDTH, BOARD_WIDTH, BOARD_HEIGHT, SET_POINT, FRAME_RATE
from pongpy.interfaces.team import Team
from pongpy.models.color import Color

logger = getLogger(__name__)


class Game:
    team1: Team
    team2: Team

    def __init__(self, team1: Team, team2: Team):
        self.team1 = team1
        self.team2 = team2

    def start(self):
        Pong(self.team1, self.team2)


class Pong:
    board: Board
    left_team: Team
    right_team: Team
    result: dict
    turn: bool
    games = []

    def __init__(self, left_team: Team, right_team: Team):
        pyxel.init(WIDTH, HEIGHT, fps=FRAME_RATE)
        self.left_team = left_team
        self.right_team = right_team
        self.board = Board(BOARD_WIDTH, BOARD_HEIGHT, PADDING, PADDING, left_team=self.left_team, right_team=self.right_team)
        pyxel.run(self.update, self.draw)

    def update(self):

        self.board.update()

        # ゲーム終了条件
        if self.board.p1.score >= SET_POINT or self.board.p2.score >= SET_POINT:
            # デュース判定
            if abs(self.board.p1.score - self.board.p2.score) >= 2:
                print(f'{self.board.p1.score_label} {self.board.p2.score_label}')
                # 最新のPyxel だと quit したときに Python 自体 exit してしまう
                pyxel.quit()

    def draw(self):
        pyxel.cls(Color.BLACK.value)
        self.board.draw()
        pyxel.text(PADDING * 2, HEIGHT - PADDING * 2, ''.join('o' if x else 'x' for x in self.games), Color.GRAY.value)



from typing import NamedTuple

from .pos import Pos


class TeamState(NamedTuple):
    """
    チーム内の現在の情報
    """
    atk_pos: Pos
    def_pos: Pos
    score: int  # 1セット内での得点数


class State(NamedTuple):
    """
    ボールの位置や自機の座標など現在の情報
    """
    is_right_side: bool  # 右側かどうか
    mine_team: TeamState  # 自チーム
    enemy_team: TeamState  # 相手チーム
    ball_pos: Pos  # ボール
    time: int  # 経過フレーム数



from pongpy.interfaces.team import Team
from pongpy.models.game_info import GameInfo
from pongpy.models.state import State


class StubTeam(Team):
    """
    初期位置のまま動かないサンプル。
    """
    @property
    def name(self) -> str:
        return 'stub team'

    def atk_action(self, info: GameInfo, state: State) -> int:
        return 0

    def def_action(self, info: GameInfo, state: State) -> int:
        return 0



import random
from typing import Tuple

import pyxel

from pongpy.controllers.ball import Ball
from pongpy.controllers.team_manager import TeamManager
from pongpy.interfaces.team import Team
from pongpy.models.color import Color
from pongpy.models.game_info import GameInfo
from pongpy.models.pos import Pos
from pongpy.models.state import State
from pongpy.definitions import BALL_SIZE, BAR_WIDTH, ATK_SIZE, DEF_SIZE, PADDING, MAX_VY


class Board:
    game_info: GameInfo
    p1: TeamManager
    p2: TeamManager
    offset_x: int
    offset_y: int
    ball: Ball

    def __init__(self, width: int, height: int, offset_x: int, offset_y: int, left_team: Team, right_team: Team):
        self.game_info = GameInfo(width=width, height=height)
        self.p1 = TeamManager(self.game_info, left_team, reversed=False)
        self.p2 = TeamManager(self.game_info, right_team, reversed=True)
        self.offset_x = offset_x
        self.offset_y = offset_y
        self.ball = Ball(BALL_SIZE, Pos(width // 2, height // 2), 1, random.random() * 5)

    def absolute(self, pos) -> Tuple[int, int]:
        return pos.x + self.offset_x, pos.y + self.offset_y

    def update(self):
        time = pyxel.frame_count
        p1_state = State(is_right_side=False, mine_team=self.p1.state, enemy_team=self.p2.state, ball_pos=self.ball.pos, time=time)
        p2_state = State(is_right_side=True, mine_team=self.p2.state, enemy_team=self.p1.state, ball_pos=self.ball.pos, time=time)
        self.p1.update(p1_state)
        self.p2.update(p2_state)

        ball_pos = self.ball.updated()
        self.ball.pos = ball_pos
        if ball_pos.y <= 0:
            self.ball.pos = Pos(ball_pos.x, -ball_pos.y)
            self.ball.vy *= -1
        if ball_pos.y >= self.game_info.height:
            self.ball.pos = Pos(ball_pos.x, self.game_info.height - (ball_pos.y - self.game_info.height))
            self.ball.vy *= -1
        if ball_pos.x < 0:
            # p2 ゴール
            self.p2.add_score()
            self.ball.pos = Pos(self.game_info.width // 2, self.game_info.height // 2)
            self.ball.vx = 1
            self.ball.vy = random.random() * 2
        if ball_pos.x > self.game_info.width:
            # p1 ゴール
            self.p1.add_score()
            self.ball.pos = Pos(self.game_info.width // 2, self.game_info.height // 2)
            self.ball.vx = -1
            self.ball.vy = random.random() * 2
        for player in (self.p1, self.p2):
            if player.atk_pos.x - BAR_WIDTH // 2 <= ball_pos.x <= player.atk_pos.x + BAR_WIDTH // 2:
                if player.atk_pos.y - ATK_SIZE // 2 <= ball_pos.y <= player.atk_pos.y + ATK_SIZE // 2:
                    self.ball.vx = self.ball.vx * -1
                    diff = ball_pos.y - player.atk_pos.y + (random.random() - 0.5) * 2
                    self.ball.vy += diff
            if player.def_pos.x - BAR_WIDTH // 2 <= ball_pos.x <= player.def_pos.x + BAR_WIDTH // 2:
                if player.def_pos.y - DEF_SIZE // 2 <= ball_pos.y <= player.def_pos.y + DEF_SIZE // 2:
                    self.ball.vx = self.ball.vx * -1
                    diff = ball_pos.y - player.def_pos.y
                    self.ball.vy += diff // 6

        # 最高速度規制
        if self.ball.vy > MAX_VY:
            self.ball.vy = MAX_VY
        if self.ball.vy < -MAX_VY:
            self.ball.vy = -MAX_VY

    def draw(self):
        pyxel.rect(
            self.offset_x,
            self.offset_y,
            self.game_info.width,
            self.game_info.height,
            Color.GRAY.value)

        ball_x, ball_y = self.absolute(self.ball.pos)
        pyxel.rect(
            ball_x - BALL_SIZE // 2,
            ball_y - BALL_SIZE // 2,
            BALL_SIZE,
            BALL_SIZE,
            Color.GREEN.value)

        for player in (self.p1, self.p2):
            x, y = self.absolute(player.atk_pos)
            pyxel.rect(
                x - BAR_WIDTH // 2,
                y - ATK_SIZE // 2,
                BAR_WIDTH,
                ATK_SIZE,
                Color.BLUE.value)

            x, y = self.absolute(player.def_pos)
            pyxel.rect(
                x - BAR_WIDTH // 2,
                y - DEF_SIZE // 2,
                BAR_WIDTH,
                DEF_SIZE,
                Color.YELLO.value)
        x, y = self.absolute(Pos(0, self.game_info.height))
        scores = f'{self.p1.score_label} {self.p2.score_label}'
        pyxel.text(x + PADDING, y + PADDING, scores, Color.GRAY.value)



from setuptools import setup, find_packages

setup(
    name="pongpy",
    url='https://github.com/pistatium/pong',
    description='Pong game',
    version="0.2.1",
    license='MIT',
    author='pistatium',
    packages=find_packages(),
    install_requires=[
        "pyxel==1.2.6",
        "click"
    ],
    entry_points={
        "gui_scripts": [
            "pongpy = pongpy.cmd:main"
        ]
    }
)



import random
from pongpy.interfaces.team import Team
from pongpy.models.game_info import GameInfo
from pongpy.models.state import State


class RandomTeam(Team):
    """
    ランダムに折り返して動くサンプル。
    """
    atk_direction = -1
    def_direction = -1

    @property
    def name(self) -> str:
        return 'random team'

    def atk_action(self, info: GameInfo, state: State) -> int:
        if random.random() > 0.9:
            self.atk_direction *= -1
        return self.atk_direction

    def def_action(self, info: GameInfo, state: State) -> int:
        if random.random() > 0.95:
            self.def_direction *= -1
        return self.def_direction



import pyxel

from pongpy.interfaces.team import Team
from pongpy.models.game_info import GameInfo
from pongpy.models.state import State


class ManualTeam(Team):
    """
    デバッグ用の手動操作チーム。
    Pyxel を直接読んでいるのでデバッグ用途以外では利用しないでください。
    """
    @property
    def name(self) -> str:
        return 'Manual'

    def atk_action(self, info: GameInfo, state: State) -> int:
        if pyxel.btn(pyxel.KEY_I):
            return -info.atk_return_limit
        if pyxel.btn(pyxel.KEY_K):
            return info.atk_return_limit
        return 0

    def def_action(self, info: GameInfo, state: State) -> int:
        if pyxel.btn(pyxel.KEY_W):
            return -info.def_return_limit
        if pyxel.btn(pyxel.KEY_S):
            return info.def_return_limit
        return 0
